/*
 * Copyright (c) 2005-2008 Obeo
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Obeo - initial API and implementation
 */

package fr.obeo.acceleo.gen.template.scripts.imports;

import java.io.File;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.TreeSet;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.Platform;
import org.eclipse.emf.ecore.EObject;
import org.osgi.framework.Bundle;

import fr.obeo.acceleo.ecore.factories.FactoryException;
import fr.obeo.acceleo.gen.AcceleoGenMessages;
import fr.obeo.acceleo.gen.template.TemplateConstants;
import fr.obeo.acceleo.gen.template.eval.ENode;
import fr.obeo.acceleo.gen.template.eval.ENodeCastException;
import fr.obeo.acceleo.gen.template.eval.ENodeException;
import fr.obeo.acceleo.gen.template.eval.ENodeList;
import fr.obeo.acceleo.gen.template.eval.LaunchManager;
import fr.obeo.acceleo.gen.template.expressions.TemplateCallExpression;
import fr.obeo.acceleo.gen.template.expressions.TemplateExpression;
import fr.obeo.acceleo.gen.template.expressions.TemplateLiteralExpression;
import fr.obeo.acceleo.gen.template.scripts.AbstractScript;
import fr.obeo.acceleo.gen.template.scripts.IEvalSettings;
import fr.obeo.acceleo.gen.template.scripts.IScript;
import fr.obeo.acceleo.gen.template.scripts.imports.services.ContextServices;
import fr.obeo.acceleo.gen.template.scripts.imports.services.ENodeServices;
import fr.obeo.acceleo.gen.template.scripts.imports.services.EObjectServices;
import fr.obeo.acceleo.gen.template.scripts.imports.services.RequestServices;
import fr.obeo.acceleo.tools.classloaders.AcceleoClassLoader;
import fr.obeo.acceleo.tools.classloaders.AcceleoGenClassLoader;
import fr.obeo.acceleo.tools.plugins.AcceleoModuleProvider;

/**
 * Java services element that can be used during a code generation.
 * 
 * @author www.obeo.fr
 * 
 */
public class EvalJavaService implements IEvalSettings {

	private static String[] SERVICES_WITH_TYPE_RESOLVE = new String[] { "filter", "cast", "until", "eContainer", "eAllContents", "current" }; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$

	private static String[] SERVICES_WITH_TYPE_BRIDGE = new String[] { "select", "delete", "sep", "nGet", "trace", "debug", "nFirst", "nLast", "sort", "nSort", "minimize", "nMinimize", "reverse", //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$ //$NON-NLS-7$ //$NON-NLS-8$ //$NON-NLS-9$ //$NON-NLS-10$ //$NON-NLS-11$ //$NON-NLS-12$ //$NON-NLS-13$
			"nReverse" }; //$NON-NLS-1$

	private static List getTypeResolveList() {
		if (typeResolveList == null) {
			typeResolveList = new ArrayList();
			for (int i = 0; i < SERVICES_WITH_TYPE_RESOLVE.length; i++) {
				typeResolveList.add(SERVICES_WITH_TYPE_RESOLVE[i]);
			}
		}
		return typeResolveList;
	}

	private static List typeResolveList = null;

	private static List getTypeBridgeList() {
		if (typeBridgeList == null) {
			typeBridgeList = new ArrayList();
			for (int i = 0; i < SERVICES_WITH_TYPE_BRIDGE.length; i++) {
				typeBridgeList.add(SERVICES_WITH_TYPE_BRIDGE[i]);
			}
		}
		return typeBridgeList;
	}

	private static List typeBridgeList = null;

	/**
	 * The default evaluation mode.
	 */
	public static final int MODE_DEFAULT = 0;

	/**
	 * The 'ENode' evaluation mode : ENode services are available.
	 */
	public static final int MODE_ENODE = 1;

	/**
	 * The 'ENodeList' evaluation mode : ENodeList services are available.
	 */
	public static final int MODE_LIST = 2;

	/**
	 * The evaluation mode.
	 */
	protected int mode = MODE_DEFAULT;

	/**
	 * Instance that contains the services.
	 */
	protected Object instance;

	/**
	 * The class loader.
	 */
	private ClassLoader loader = null;

	/**
	 * Gets the service for the given name
	 */
	private Map name2service = new HashMap();

	/**
	 * Indicates if the services have a script context.
	 */
	private boolean hasScriptContext = true;

	/**
	 * Constructor.
	 * 
	 * @param instance
	 *            is the instance that contains the services
	 * @param hasScriptContext
	 *            indicates if the services have a script context
	 */
	public EvalJavaService(Object instance, boolean hasScriptContext) {
		this.instance = instance;
		this.hasScriptContext = hasScriptContext;
		initializeName2service();
	}

	/**
	 * Constructor.
	 * 
	 * @param script
	 *            is the script that imports the service
	 * @param className
	 *            is the class to be created
	 * @throws JavaServiceNotFoundException
	 */
	public EvalJavaService(File script, String className) throws JavaServiceNotFoundException {
		// Remark : script "fr.package.template" -> class "fr.package.Template"
		String shortName = className;
		String path = ""; //$NON-NLS-1$
		int jDot = className.lastIndexOf("."); //$NON-NLS-1$
		if (jDot > -1) {
			shortName = className.substring(jDot + 1);
			path = className.substring(0, jDot);
		}
		if (shortName.length() > 0)
			shortName = shortName.substring(0, 1).toUpperCase() + shortName.substring(1);
		className = ((path.length() > 0) ? path + '.' : "") + shortName; //$NON-NLS-1$
		initialize(script, className);
		initializeName2service();
	}

	/**
	 * Constructor.
	 * 
	 * @param script
	 *            is the script that imports the service
	 * @throws JavaServiceNotFoundException
	 */
	public EvalJavaService(File script) throws JavaServiceNotFoundException {
		// Remark : script "/fr/package/template.mt" -> class
		// "fr.package.Template"
		String name = new Path(script.getName()).removeFileExtension().lastSegment();
		if (name.length() > 0)
			name = Character.toUpperCase(name.charAt(0)) + name.substring(1);
		String[] segments = getPackagePath(script).append(name).segments();
		// ASSERT (segments.length > 0){
		// Java class name
		StringBuffer className = new StringBuffer();
		className.append(segments[0]);
		for (int i = 1; i < segments.length; i++) {
			className.append('.');
			className.append(segments[i]);
		}
		initialize(script, className.toString());
		initializeName2service();
	}

	/**
	 * @return the evaluation mode
	 */
	public int getMode() {
		return mode;
	}

	/**
	 * @param mode
	 *            is the evaluation mode
	 */
	public void setMode(int mode) {
		this.mode = mode;
	}

	/**
	 * @return true if the services have a script context
	 */
	public boolean hasScriptContext() {
		return hasScriptContext;
	}

	/**
	 * @return the instance
	 */
	public Object getInstance() {
		return instance;
	}

	private void initialize(File script, String className) throws JavaServiceNotFoundException {
		try {
			loader = createClassLoader(script);
			Class c = loader.loadClass(className);
			instance = c.newInstance();
		} catch (Exception e) {
			throw new JavaServiceNotFoundException(AcceleoGenMessages.getString("EvalJavaService.UnresolvedService", new Object[] { e.getMessage(), })); //$NON-NLS-1$
		}
	}

	private void initializeName2service() {
		name2service.clear();
		if (instance != null) {
			Method[] methods = instance.getClass().getMethods();
			for (int i = 0; i < methods.length; i++) {
				if (methods[i].getDeclaringClass() == instance.getClass() && methods[i].getParameterTypes().length > 0) {
					String key = methods[i].getName() + methods[i].getParameterTypes().length;
					Method[] name2values = (Method[]) name2service.get(key);
					if (name2values == null) {
						name2service.put(key, new Method[] { methods[i] });
					} else {
						List name2values_ = new ArrayList(name2values.length + 1);
						boolean ok = false;
						for (int j = 0; j < name2values.length; j++) {
							if (name2values[j].getParameterTypes()[0].isAssignableFrom(methods[i].getParameterTypes()[0]) || name2values[j].getParameterTypes()[0] == ENode.class) {
								name2values_.add(methods[i]);
								ok = true;
							}
							name2values_.add(name2values[j]);
						}
						if (!ok) {
							name2values_.add(methods[i]);
						}
						name2service.put(key, name2values_.toArray(new Method[name2values_.size()]));
					}
				}
			}
		}
	}

	private Method[] getPotentialMethods(String name, int argsCount) {
		String key = name + argsCount;
		return (Method[]) name2service.get(key);
	}

	/**
	 * Gets the package path of the given script
	 * 
	 * @param script
	 *            is the script
	 * @return the full path of the package
	 */
	private IPath getPackagePath(File script) {
		IFile file = ResourcesPlugin.getWorkspace().getRoot().getFileForLocation(new Path(script.getAbsolutePath()));
		if (file != null && file.isAccessible()) {
			return file.getProjectRelativePath().removeLastSegments(1).removeFirstSegments(1);
		}
		String path = AcceleoModuleProvider.getDefault().getRelativePath(script);
		if (path != null) {
			return new Path(path).removeLastSegments(1);
		} else {
			return new Path(""); //$NON-NLS-1$
		}
	}

	/**
	 * Creates the default class loader for the given script.
	 * 
	 * @param script
	 *            is the script
	 * @return the default class loader
	 */
	private ClassLoader createClassLoader(File script) {
		IFile file = ResourcesPlugin.getWorkspace().getRoot().getFileForLocation(new Path(script.getAbsolutePath()));
		if (file != null && file.isAccessible()) {
			return new AcceleoGenClassLoader(file.getProject(), EvalJavaService.class.getClassLoader());
		}
		String pluginId = AcceleoModuleProvider.getDefault().getPluginId(script);
		if (pluginId != null) {
			Bundle bundle = Platform.getBundle(pluginId);
			if (bundle != null) {
				return new AcceleoGenClassLoader(bundle, EvalJavaService.class.getClassLoader());
			}
		}
		return EvalJavaService.class.getClassLoader();
	}

	/* (non-Javadoc) */
	public ENode eGet(TemplateCallExpression call, ENode node, ENode[] args, LaunchManager mode, boolean recursiveSearch) throws FactoryException, ENodeException {
		if (name2service.containsKey(call.getLink() + (args.length + 1))) {
			if (node.isEObject()) {
				ClassLoader old = AcceleoClassLoader.getPreferredClassLoader();
				try {
					AcceleoClassLoader.setPreferredLoader(node.getEObject());
					return eGetSub(call, node, args);
				} catch (ENodeCastException e) {
					// Never catch
					return null;
				} finally {
					AcceleoClassLoader.setPreferredClassLoader(old);
				}
			} else {
				return eGetSub(call, node, args);
			}
		} else {
			return null;
		}
	}

	private ENode eGetSub(TemplateCallExpression call, ENode node, ENode[] args) throws FactoryException, ENodeException {
		// ASSERT (instance != null){
		// Get the type and the value for each argument
		List argTypesList = new ArrayList(args.length + 1);
		List argValuesList = new ArrayList(args.length + 1);
		argTypesList.add(null);
		argValuesList.add(null);
		for (int i = 0; i < args.length; i++) {
			ENode arg = args[i];
			argTypesList.add(arg.getTypeClass());
			argValuesList.add(arg.getValue());
		}
		Class[] argTypes = (Class[]) argTypesList.toArray(new Class[argTypesList.size()]);
		Object[] argValues = (Object[]) argValuesList.toArray();
		// Get the method
		Method m = null;
		if (mode == MODE_DEFAULT && !node.isNull()) {
			argTypes[0] = node.getTypeClass();
			argValues[0] = node.getValue();
			m = eGetMethod(call.getLink(), argTypes, argValues[0]);
		} else if (mode == MODE_ENODE) {
			argTypes[0] = ENode.class;
			argValues[0] = node;
			m = eGetMethod(call.getLink(), argTypes, argValues[0]);
		} else if (mode == MODE_LIST && node.isList()) {
			try {
				argTypes[0] = ENodeList.class;
				argValues[0] = node.getList();
				m = eGetMethod(call.getLink(), argTypes, argValues[0]);
			} catch (ENodeCastException e1) {
				m = null;
			}
			if (m == null) {
				try {
					argTypes[0] = List.class;
					argValues[0] = node.getList().asList();
					m = eGetMethod(call.getLink(), argTypes, argValues[0]);
				} catch (ENodeCastException e2) {
					m = null;
				}
			}
		}
		if (m == null) {
			// Try to adapt parameters
			if (instance != null) {
				Method[] methods = getPotentialMethods(call.getLink(), argTypes.length);
				if (methods != null) {
					for (int i = 0; m == null && i < methods.length; i++) {
						boolean ok = true;
						Class[] parameterTypes = methods[i].getParameterTypes();
						if (parameterTypes.length > 0) {
							if (mode == MODE_DEFAULT) {
								Class parameterType = ENode.getAdapterType(parameterTypes[0]);
								if (parameterType != null) {
									argTypes[0] = parameterType;
									try {
										argValues[0] = node.getAdapterValue(parameterType);
									} catch (ENodeCastException e) {
										ok = false;
									}
								} else {
									ok = false;
								}
							}
							if ((mode == MODE_DEFAULT && ok) || (mode == MODE_ENODE && parameterTypes[0] == ENode.class) || (mode == MODE_LIST && parameterTypes[0] == ENodeList.class)) {
								for (int j = 1; j < parameterTypes.length; j++) {
									Class parameterType = ENode.getAdapterType(parameterTypes[j]);
									if (parameterType != null) {
										argTypes[j] = parameterType;
										try {
											argValues[j] = args[j - 1].getAdapterValue(parameterType);
										} catch (ENodeCastException e) {
											ok = false;
											break;
										}
									} else {
										ok = false;
										break;
									}
								}
								if (ok) {
									m = methods[i];
								}
							}
						}
					}
				}
			}
		}
		if (m != null) {
			// Display method
			String displayString = m.getName() + " ("; //$NON-NLS-1$
			Class[] paramTypes = m.getParameterTypes();
			for (int j = 1; j < paramTypes.length; j++) { // The first
				// parameter is
				// ignored
				Class paramType = paramTypes[j];
				displayString += getSimpleName(paramType);
				if (j + 1 < paramTypes.length) {
					displayString += ", "; //$NON-NLS-1$
				}
			}
			displayString += ')';
			if (m.getReturnType() != null) {
				displayString += ' ' + getSimpleName(m.getReturnType());
			}
			displayString += " - " + getSimpleName(m.getDeclaringClass()); //$NON-NLS-1$
			// Invoke method on instance
			Object result;
			try {
				if (call.getLink().equals("select") && call.getScript() instanceof AbstractScript && call.countArguments() > 0) { //$NON-NLS-1$
					((AbstractScript) call.getScript()).contextPush(IScript.ARGUMENT_POSITION, call.getFirstArgument().getPos());
				}
				try {
					// void => ""
					if (m.getReturnType() != null && "void".equals(getSimpleName(m.getReturnType()))) { //$NON-NLS-1$
						m.invoke(instance, argValues);
						result = ""; //$NON-NLS-1$
					} else {
						result = m.invoke(instance, argValues);
					}
				} catch (Exception npe) {
					if (mode == MODE_DEFAULT && argValues.length > 0 && argValues[0] == null) {
						result = null;
					} else {
						throw npe;
					}
				} finally {
					if (call.getLink().equals("select") && call.getScript() instanceof AbstractScript && call.countArguments() > 0) { //$NON-NLS-1$
						((AbstractScript) call.getScript()).contextPop(IScript.ARGUMENT_POSITION);
					}
				}
			} catch (InvocationTargetException e) {
				if (e.getTargetException() instanceof FactoryException) {
					throw (FactoryException) e.getTargetException();
				} else if (e.getTargetException() instanceof ENodeException) {
					throw (ENodeException) e.getTargetException();
				} else {
					final StringBuffer errorMessage = new StringBuffer("\n"); //$NON-NLS-1$
					errorMessage.append(((e.getTargetException() != null) ? e.getTargetException().getClass().getName() + " : " + e.getTargetException().getMessage() : "")); //$NON-NLS-1$ //$NON-NLS-2$
					throw new ENodeException(
							AcceleoGenMessages.getString("EvalJavaService.RuntimeException", new Object[] { displayString, errorMessage.toString(), }), call.getPos(), call.getScript(), //$NON-NLS-1$
							node, true, e.getTargetException());
				}
			} catch (Exception e) {
				final StringBuffer errorMessage = new StringBuffer("\n"); //$NON-NLS-1$
				errorMessage.append(e.getClass().getName());
				errorMessage.append(" : "); //$NON-NLS-1$
				errorMessage.append(e.getMessage());
				throw new ENodeException(
						AcceleoGenMessages.getString("EvalJavaService.RuntimeException", new Object[] { displayString, errorMessage.toString(), }), call.getPos(), call.getScript(), node, true, //$NON-NLS-1$
						e);
			}
			ENode createTry = ENode.createTry(result, node);
			if (createTry != null) {
				return createTry;
			} else {
				throw new ENodeException(
						AcceleoGenMessages.getString("EvalJavaService.RuntimeType", new Object[] { displayString, result.getClass().getName(), }), call.getPos(), call.getScript(), node, true); //$NON-NLS-1$
			}
		} else {
			// Service doesn't exist
			return null;
		}
	}

	/**
	 * Gets the simple name of the given class.
	 * <p>
	 * Sample : return 'Object' for 'java.lang.Object'
	 * 
	 * @param c
	 *            is the class
	 * @return the name
	 */
	public static String getSimpleName(Class c) {
		String name = c.getName();
		int i = name.lastIndexOf("."); //$NON-NLS-1$
		if (i > -1) {
			name = name.substring(i + 1);
		}
		return name;
	}

	private Method eGetMethod(String name, Class[] argTypes, Object receiver) {
		Method[] methods = getPotentialMethods(name, argTypes.length);
		if (methods != null) {
			for (int i = 0; i < methods.length; i++) {
				Class[] parameterTypes = methods[i].getParameterTypes();
				// ASSERT parameterTypes.length > 0
				if (argTypes[0] != EObject.class || parameterTypes[0].isInstance(receiver)) {
					boolean ok = true;
					for (int j = 0; j < parameterTypes.length; j++) {
						if ((argTypes[j] == EObject.class && !argTypes[j].isAssignableFrom(parameterTypes[j])) || (argTypes[j] != EObject.class && argTypes[j] != parameterTypes[j])) {
							ok = false;
							break;
						}
					}
					if (ok) {
						return methods[i];
					}
				}
			}
		}
		return null;
	}

	/* (non-Javadoc) */
	public Object resolveType(Object type, TemplateCallExpression call, int depth) {
		if (instance != null && type != null) { // Remark : type == GENERIC_TYPE
			// => type != null
			List quickKey = new ArrayList();
			quickKey.add(type);
			quickKey.add(call.toString()); // not getLink() because arguments
			if (quickResolveType.containsKey(quickKey)) {
				return quickResolveType.get(quickKey);
			} else {
				Object resolvedType = null;
				// 'filter' case
				if (instance instanceof ENodeServices || instance instanceof EObjectServices) {
					if (getTypeResolveList().contains(call.getLink()) && call.countArguments() > 0) {
						TemplateExpression firstArg = call.getFirstArgument();
						if (firstArg instanceof TemplateLiteralExpression) {
							Object result = ((TemplateLiteralExpression) firstArg).resolveAsEClassifier();
							if (result != null) {
								resolvedType = result;
							}
						}
					}
				}
				if (resolvedType == null && (instance instanceof RequestServices || instance instanceof ENodeServices || instance instanceof ContextServices)) {
					if (getTypeBridgeList().contains(call.getLink())) {
						resolvedType = type;
					}
				}
				// Use declared methods
				if (resolvedType == null) {
					Method[] methods = getPotentialMethods(call.getLink(), call.countArguments() + 1);
					if (methods != null && methods.length > 0 && methods[0].getDeclaringClass() == instance.getClass()) {
						resolvedType = GENERIC_TYPE;
					}
				}
				quickResolveType.put(quickKey, resolvedType);
				return resolvedType;
			}
		} else {
			return null;
		}
	}

	private Map quickResolveType = new HashMap();

	/* (non-Javadoc) */
	public Object[] getCompletionProposals(Object type, int depth) {
		if (instance != null) {
			Collection result = new TreeSet(new Comparator() {
				public int compare(Object arg0, Object arg1) {
					Method m0 = ((Method) arg0);
					Method m1 = ((Method) arg1);
					return (m0.getName() + m0.toString()).compareTo(m1.getName() + m1.toString());
				}
			});
			Method[] methods = instance.getClass().getMethods();
			for (int i = 0; i < methods.length; i++) {
				Method method = methods[i];
				if (method.getDeclaringClass() == instance.getClass() && method.getParameterTypes().length >= 1) {
					result.add(method);
				}
			}
			return result.toArray();
		} else {
			return new Object[] {};
		}
	}

	/* (non-Javadoc) */
	public boolean validateCall(TemplateCallExpression call) {
		return "".equals(call.getPrefix()) || TemplateConstants.LINK_PREFIX_JAVA.equals(call.getPrefix()); //$NON-NLS-1$
	}

}
