/*
 * Copyright (c) 2005-2008 Obeo
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Obeo - initial API and implementation
 */

package fr.obeo.acceleo.gen.ui.editors.template;

import java.io.File;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IResourceChangeEvent;
import org.eclipse.core.resources.IResourceChangeListener;
import org.eclipse.core.resources.IResourceDelta;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.jdt.internal.ui.JavaPlugin;
import org.eclipse.jdt.ui.PreferenceConstants;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.preference.IPreferenceStore;
import org.eclipse.jface.text.IDocument;
import org.eclipse.jface.text.source.ISourceViewer;
import org.eclipse.jface.text.source.SourceViewerConfiguration;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.ISelectionChangedListener;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.SelectionChangedEvent;
import org.eclipse.swt.custom.StyledText;
import org.eclipse.ui.IEditorInput;
import org.eclipse.ui.IFileEditorInput;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.editors.text.TextEditor;
import org.eclipse.ui.texteditor.ChainedPreferenceStore;
import org.eclipse.ui.texteditor.IDocumentProvider;
import org.eclipse.ui.texteditor.SourceViewerDecorationSupport;
import org.eclipse.ui.views.contentoutline.IContentOutlinePage;

import fr.obeo.acceleo.gen.AcceleoEcoreGenPlugin;
import fr.obeo.acceleo.gen.template.Template;
import fr.obeo.acceleo.gen.template.TemplateConstants;
import fr.obeo.acceleo.gen.template.TemplateSyntaxExceptions;
import fr.obeo.acceleo.gen.template.expressions.TemplateExpression;
import fr.obeo.acceleo.gen.template.scripts.ScriptDescriptor;
import fr.obeo.acceleo.gen.template.scripts.SpecificScript;
import fr.obeo.acceleo.gen.template.statements.TemplateFeatureStatement;
import fr.obeo.acceleo.gen.template.statements.TemplateForStatement;
import fr.obeo.acceleo.gen.template.statements.TemplateIfStatement;
import fr.obeo.acceleo.gen.ui.AcceleoEcoreGenUiPlugin;
import fr.obeo.acceleo.gen.ui.AcceleoGenUIMessages;
import fr.obeo.acceleo.gen.ui.builders.AcceleoGenCompileOperation;
import fr.obeo.acceleo.gen.ui.editors.reflective.IGeneratorProvider;
import fr.obeo.acceleo.tools.resources.Resources;
import fr.obeo.acceleo.tools.strings.Int2;
import fr.obeo.acceleo.tools.strings.TextSearch;

/**
 * The standard text editor for template file resources (IFile).
 * 
 * @author www.obeo.fr
 * 
 */
public class AcceleoTemplateEditor extends TextEditor implements IGeneratorProvider, IResourceChangeListener {

	/**
	 * The color manager for the template editor.
	 */
	protected ColorManager colorManager = new ColorManager();

	/**
	 * Template content outline page.
	 */
	protected AcceleoTemplateOutlinePage contentOutlinePage = null;

	/**
	 * A listener which is notified when the outline's selection changes.
	 */
	protected ISelectionChangedListener selectionChangedListener;

	/**
	 * Specific script model.
	 */
	protected SpecificScript generator = null;

	/**
	 * Compares the current input of the editor with the last saved input.
	 */
	protected LastSaveComparator comparator = new LastSaveComparator(this);

	/** The editor's blocks matcher */
	protected AcceleoPairMatcher blockMatcher = new AcceleoPairMatcher();

	/** Preference key for matching brackets */
	protected final static String MATCHING_BRACKETS = PreferenceConstants.EDITOR_MATCHING_BRACKETS;

	/** Preference key for matching brackets color */
	protected final static String MATCHING_BRACKETS_COLOR = PreferenceConstants.EDITOR_MATCHING_BRACKETS_COLOR;

	/**
	 * Constructor.
	 */
	public AcceleoTemplateEditor() {
		super();
	}

	/* (non-Javadoc) */
	public SpecificScript getGenerator() {
		return generator;
	}

	/**
	 * @return the source viewer
	 */
	public ISourceViewer sourceViewer() {
		return getSourceViewer();
	}

	/* (non-Javadoc) */
	protected void doSetInput(IEditorInput input) throws CoreException {
		IFile file = (input instanceof IFileEditorInput) ? ((IFileEditorInput) input).getFile() : null;
		String content = Resources.getFileContent(file).toString();
		TemplateConstants.initConstants(content);
		setSourceViewerConfiguration(createSourceViewerConfiguration());
		setDocumentProvider(createDocumentProvider());
		super.doSetInput(input);
	}

	/* (non-Javadoc) */
	protected void initializeKeyBindingScopes() {
		setKeyBindingScopes(new String[] { "fr.obeo.acceleo.gen.template.editor" }); //$NON-NLS-1$
	}

	/**
	 * Creates the source viewer configuration.
	 * 
	 * @return the source viewer configuration
	 */
	protected SourceViewerConfiguration createSourceViewerConfiguration() {
		return new AcceleoTemplateConfiguration(this, colorManager);
	}

	/**
	 * Creates the document provider.
	 * 
	 * @return the document provider
	 */
	protected IDocumentProvider createDocumentProvider() {
		return new AcceleoTemplateDocumentProvider(colorManager);
	}

	/* (non-Javadoc) */
	public void dispose() {
		super.dispose();
		colorManager.dispose();
		getContentOutlinePage().removeSelectionChangedListener(selectionChangedListener);
		ResourcesPlugin.getWorkspace().removeResourceChangeListener(this);
		/*
		 * Dispose the block matcher
		 */
		if (blockMatcher != null) {
			blockMatcher.dispose();
			blockMatcher = null;
		}
	}

	/* (non-Javadoc) */
	public Object getAdapter(Class type) {
		if (type.equals(IContentOutlinePage.class)) {
			return getContentOutlinePage();
		}
		return super.getAdapter(type);
	}

	/**
	 * Returns the template content outline page. Creates the listener which is
	 * notified when the outline's selection changes.
	 * 
	 * @return the template content outline page
	 */
	protected AcceleoTemplateOutlinePage getContentOutlinePage() {
		if (contentOutlinePage == null) {
			contentOutlinePage = createContentOutlinePage();
			selectionChangedListener = createSelectionChangeListener();
			contentOutlinePage.addSelectionChangedListener(selectionChangedListener);
			update();
			ResourcesPlugin.getWorkspace().addResourceChangeListener(this);
		}
		return contentOutlinePage;
	}

	/* (non-Javadoc) */
	public void resourceChanged(final IResourceChangeEvent event) {
		if (PlatformUI.getWorkbench().getActiveWorkbenchWindow() != null && PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage() != null
				&& PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage().getActiveEditor() != this) {
			if (event.getType() == IResourceChangeEvent.POST_CHANGE && deltaResources(event.getDelta()).contains(getFile())) {
				try {
					init(getEditorSite(), getEditorInput());
				} catch (PartInitException e) {
					AcceleoEcoreGenUiPlugin.getDefault().log(e, true);
				}
			}
		}
	}

	private List deltaResources(IResourceDelta delta) {
		List files = new ArrayList();
		IResource resource = delta.getResource();
		if (resource instanceof IFile) {
			if (delta.getKind() == IResourceDelta.CHANGED) {
				files.add(resource);
			}
		}
		IResourceDelta[] children = delta.getAffectedChildren();
		for (int i = 0; i < children.length; i++) {
			files.addAll(deltaResources(children[i]));
		}
		return files;
	}

	/**
	 * Creates the content outline page.
	 * 
	 * @return the content outline page
	 */
	protected AcceleoTemplateOutlinePage createContentOutlinePage() {
		return new AcceleoTemplateOutlinePage(this);
	}

	/**
	 * Creates a listener which is notified when the outline's selection
	 * changes.
	 * 
	 * @return the listener which is notified when the outline's selection
	 *         changes
	 */
	protected ISelectionChangedListener createSelectionChangeListener() {
		return new ISelectionChangedListener() {
			public void selectionChanged(SelectionChangedEvent event) {
				selectionChangedDetected(event);
			}
		};
	}

	/* non-javadoc */
	protected void configureSourceViewerDecorationSupport(SourceViewerDecorationSupport support) {
		/*
		 * Set the block matcher
		 */
		support.setCharacterPairMatcher(blockMatcher);
		support.setMatchingCharacterPainterPreferenceKeys(MATCHING_BRACKETS, MATCHING_BRACKETS_COLOR);

		IPreferenceStore pref = JavaPlugin.getDefault().getPreferenceStore();
		IPreferenceStore[] stores = { getPreferenceStore(), pref };
		setPreferenceStore(new ChainedPreferenceStore(stores));
		support.install(getPreferenceStore());
		super.configureSourceViewerDecorationSupport(support);
	}

	/**
	 * Methods which is notified when the outline's selection changes.
	 * 
	 * @param event
	 *            is the selection changed event
	 */
	protected void selectionChangedDetected(SelectionChangedEvent event) {
		ISelection selection = event.getSelection();
		Object selectedElement = ((IStructuredSelection) selection).getFirstElement();
		if (selectedElement instanceof Map.Entry) {
			String selectedName = ((ScriptDescriptor) ((Map.Entry) selectedElement).getKey()).getName();
			String selectedType = ((ScriptDescriptor) ((Map.Entry) selectedElement).getKey()).getType();
			// Search script declaration, give the good result even when the
			// text has been modified
			String text = getSourceViewer().getDocument().get();
			String templateType = null;
			String templateName = null;
			TemplateConstants.initConstants(text);
			Int2 end = new Int2(0, 0);
			while (end.e() > -1 && end.e() < text.length()) {
				Int2 begin = TextSearch.getDefaultSearch().indexOf(text, TemplateConstants.SCRIPT_BEGIN, end.e(), null, TemplateConstants.INHIBS_SCRIPT_CONTENT);
				if (begin.b() > -1) {
					// Clear informations
					templateType = null;
					templateName = null;
					// Prepare new script
					end = TextSearch.getDefaultSearch().blockIndexEndOf(text, TemplateConstants.SCRIPT_BEGIN, TemplateConstants.SCRIPT_END, begin.b(), false, TemplateConstants.SPEC,
							TemplateConstants.INHIBS_SCRIPT_DECLA);
					if (end.e() > -1) {
						// Information's positions
						boolean error = false;
						Int2[] properties = TextSearch.getDefaultSearch().splitPositionsIn(text, begin.e(), end.b(), TemplateConstants.SCRIPT_PROPERTIES_SEPARATORS, false, TemplateConstants.SPEC,
								TemplateConstants.INHIBS_SCRIPT_DECLA);
						for (int i = 0; !error && i < properties.length; i += 2) {
							String key = text.substring(properties[i].b(), properties[i].e());
							if ((i + 1) < properties.length) {
								String value = text.substring(properties[i + 1].b(), properties[i + 1].e());
								if (value.length() >= 2 && value.startsWith(TemplateConstants.LITERAL[0]) && value.endsWith(TemplateConstants.LITERAL[1])) {
									value = value.substring(1, value.length() - 1);
									if (key.equals(TemplateConstants.SCRIPT_TYPE)) {
										templateType = value;
									} else if (key.equals(TemplateConstants.SCRIPT_NAME)) {
										templateName = value;
									} else {
										error = true;
									}
								} else {
									error = true;
								}
							} else {
								error = true;
							}
						}
					}
					// This declaration is corresponding to the selected
					// element?
					if (selectedType.equals(templateType) && selectedName.equals(templateName)) {
						selectRange(begin.b(), end.e());
					}
				} else { // -1
					end = new Int2(text.length(), text.length());
				}
			}
		} else if (selectedElement instanceof TemplateIfStatement) {
			Int2 pos = ((TemplateIfStatement) selectedElement).getCondition().getPos();
			pos = comparator.save2currentPosition(new Int2(pos.b() - TemplateConstants.IF_BEGIN.length(), pos.e() + TemplateConstants.IF_THEN.length()));
			selectRange(pos.b(), pos.e());
		} else if (selectedElement instanceof TemplateForStatement) {
			Int2 pos = ((TemplateForStatement) selectedElement).getCondition().getPos();
			pos = comparator.save2currentPosition(new Int2(pos.b() - TemplateConstants.FOR_BEGIN.length(), pos.e() + TemplateConstants.FOR_THEN.length()));
			selectRange(pos.b(), pos.e());
		} else if (selectedElement instanceof TemplateFeatureStatement) {
			Int2 pos = ((TemplateFeatureStatement) selectedElement).getPos();
			pos = comparator.save2currentPosition(new Int2(pos.b() - TemplateConstants.FEATURE_BEGIN.length(), pos.e() + TemplateConstants.FEATURE_END.length()));
			selectRange(pos.b(), pos.e());
		} else if (selectedElement instanceof TemplateExpression) {
			Int2 pos = ((TemplateExpression) selectedElement).getPos();
			pos = comparator.save2currentPosition(new Int2(pos.b(), pos.e()));
			selectRange(pos.b(), pos.e());
		} else if (selectedElement instanceof Template) {
			Int2 pos = comparator.save2currentPosition(((Template) selectedElement).getPos());
			selectRange(pos.b(), pos.e());
		}
	}

	/**
	 * Sets the highlighted range of this text editor to the specified region.
	 * 
	 * @param begin
	 *            is the beginning index
	 * @param end
	 *            is the ending index
	 */
	protected void selectRange(int begin, int end) {
		if (begin > -1 && end >= begin) {
			ISourceViewer viewer = getSourceViewer();
			StyledText widget = viewer.getTextWidget();
			widget.setRedraw(false);
			try {
				setHighlightRange(begin, end - begin, true);
			} catch (Exception e) {
			}
			selectAndReveal(begin, end - begin);
			widget.setRedraw(true);
		}
	}

	/* (non-Javadoc) */
	public void doSave(IProgressMonitor progressMonitor) {
		firstGeneratorToBuild = getFile();
		AcceleoGenCompileOperation.getFile2ScriptMap().remove(firstGeneratorToBuild.getFullPath().toString());
		super.doSave(progressMonitor);
		update();
	}

	/**
	 * @return the first generator to build
	 */
	public static IFile getFirstGeneratorToBuild() {
		return firstGeneratorToBuild;
	}

	private static IFile firstGeneratorToBuild = null;

	/* (non-Javadoc) */
	public void doSaveAs() {
		super.doSaveAs();
		update();
	}

	/**
	 * It updates the generator with the editor's source viewer document.
	 */
	public void update() {
		IDocumentProvider provider = getDocumentProvider();
		if (provider != null) {
			IDocument document = provider.getDocument(getEditorInput());
			if (document != null) {
				comparator.markAsSave();
				String content = document.get();
				if (content != null) {
					String trim = content.trim();
					if (trim.length() > 0 && !trim.startsWith("<%") && !trim.startsWith("[%")) { //$NON-NLS-1$ //$NON-NLS-2$
						if (PlatformUI.getWorkbench() != null && PlatformUI.getWorkbench().getActiveWorkbenchWindow() != null) {
							MessageDialog.openWarning(PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(), AcceleoGenUIMessages.getString("AcceleoTemplateEditor.ReadOnlyWarning.Title"), //$NON-NLS-1$
									AcceleoGenUIMessages.getString("AcceleoTemplateEditor.ReadOnlyWarning.Message")); //$NON-NLS-1$
						}
					}
				}
				if (content != null) {
					if (this.generator == null) {
						IFile file = (getEditorInput() instanceof IFileEditorInput) ? ((IFileEditorInput) getEditorInput()).getFile() : null;
						this.generator = createGenerator((file != null) ? file.getLocation().toFile() : null);
						// this.generator.reset();
					}
					try {
						this.generator.reset(content);
					} catch (TemplateSyntaxExceptions e) {
					}
					getContentOutlinePage().setInput(generator);
				}
			}
		}
	}

	/**
	 * Creates the generator for the given file.
	 * 
	 * @param file
	 *            is the file
	 * @return the new generator
	 */
	protected SpecificScript createGenerator(File file) {
		SpecificScript result = AcceleoEcoreGenPlugin.getDefault().getGlobalScriptContext().getScript(file, null);
		if (result == null) {
			result = new SpecificScript(file, null, AcceleoEcoreGenPlugin.getDefault().getGlobalScriptContext());
		}
		return result;
	}

	/**
	 * Get the input file for this editor.
	 * 
	 * @return the editor input file
	 */
	public IFile getFile() {
		return (IFile) getEditorInput().getAdapter(IFile.class);
	}

}
