/*
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    David Bonneau and Obeo - initial API and implementation
 */

package fr.obeo.acceleo.gen.debug.ui.model;

import java.io.File;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IWorkspaceRunnable;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.debug.core.DebugPlugin;
import org.eclipse.debug.core.IBreakpointManager;
import org.eclipse.debug.core.model.IBreakpoint;
import org.eclipse.debug.core.model.LineBreakpoint;

import fr.obeo.acceleo.gen.debug.ui.AcceleoGenDebugUIMessages;

/**
 * Acceleo line breakpoint.
 */
public class AcceleoLineBreakpoint extends LineBreakpoint {

	/**
	 * Default constructor is required for the breakpoint manager to re-create
	 * persisted breakpoints. After instantiating a breakpoint, the
	 * <code>setMarker(...)</code> method is called to restore this breakpoint's
	 * attributes.
	 */
	public AcceleoLineBreakpoint() {
	}

	/**
	 * Constructs a line breakpoint on the given resource at the given line
	 * number. The line number is 1-based (i.e. the first line of a file is line
	 * number 1). The Acceleo VM uses 0-based line numbers, so this line number
	 * translation is done at breakpoint install time.
	 * 
	 * @param resource
	 *            file on which to set the breakpoint
	 * @param lineNumber
	 *            1-based line number of the breakpoint
	 * @throws CoreException
	 *             if unable to create the breakpoint
	 */
	public AcceleoLineBreakpoint(final IResource resource, final int lineNumber) throws CoreException {
		IWorkspaceRunnable runnable = new IWorkspaceRunnable() {
			public void run(IProgressMonitor monitor) throws CoreException {
				IMarker marker = resource.createMarker("fr.obeo.acceleo.gen.debug.ui.acceleo.markerType.lineBreakpoint"); //$NON-NLS-1$
				setMarker(marker);
				marker.setAttribute(IBreakpoint.ENABLED, Boolean.TRUE);
				marker.setAttribute(IMarker.LINE_NUMBER, lineNumber);
				marker.setAttribute(IBreakpoint.ID, getModelIdentifier());
				marker.setAttribute(IMarker.MESSAGE, AcceleoGenDebugUIMessages.getString("AcceleoLineBreakpoint.HoverText", new Object[] { resource.getName(), Integer.toString(lineNumber), })); //$NON-NLS-1$
			}
		};
		run(getMarkerRule(resource), runnable);
	}

	/* (non-Javadoc) */
	public String getModelIdentifier() {
		return IAcceleoDebugConstants.ID_ACCELEO_DEBUG_MODEL;
	}

	/**
	 * Returns the current template file.
	 * 
	 * @return the current template file
	 */
	public File getFile() {
		File ret = null;
		if (getMarker().getResource() instanceof IFile) {
			ret = ((IFile) getMarker().getResource()).getLocation().toFile();
		}
		return ret;
	}

	/* (non-Javadoc) */
	public void setEnabled(boolean enabled) throws CoreException {
		super.setEnabled(enabled);
		// Fire event for AcceleoDebutTarget
		IBreakpointManager breakpointManager = DebugPlugin.getDefault().getBreakpointManager();
		if (breakpointManager != null) {
			breakpointManager.fireBreakpointChanged(this);
		}
	}

}
