/*
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    David Bonneau and Obeo - initial API and implementation
 */

package fr.obeo.acceleo.gen.debug.ui.model;

import java.io.File;
import java.util.HashMap;
import java.util.Map;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IExtension;
import org.eclipse.core.runtime.IExtensionPoint;
import org.eclipse.core.runtime.IExtensionRegistry;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.Platform;
import org.eclipse.debug.core.DebugException;
import org.eclipse.debug.core.model.ILineBreakpoint;
import org.eclipse.debug.core.model.IValue;
import org.eclipse.debug.ui.IDebugModelPresentation;
import org.eclipse.debug.ui.IValueDetailListener;
import org.eclipse.jface.viewers.LabelProvider;
import org.eclipse.swt.graphics.Image;
import org.eclipse.ui.IEditorDescriptor;
import org.eclipse.ui.IEditorInput;
import org.eclipse.ui.IWorkbench;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.part.FileEditorInput;

import fr.obeo.acceleo.gen.debug.ui.AcceleoGenDebugUIMessages;
import fr.obeo.acceleo.gen.debug.ui.AcceleoGenDebugUiPlugin;

/**
 * Renders Acceleo debug elements.
 */
public class AcceleoModelPresentation extends LabelProvider implements IDebugModelPresentation {

	/**
	 * The debug editor extension point.
	 */
	private final static String DEBUG_EDITOR_EXTENSION_ID = "fr.obeo.acceleo.gen.debug.ui.debugeditor"; //$NON-NLS-1$

	/* (non-Javadoc) */
	public void setAttribute(String attribute, Object value) {
	}

	/* (non-Javadoc) */
	public Image getImage(Object element) {
		if (element instanceof AcceleoVariable) {
			try {
				return ((AcceleoValue) ((AcceleoVariable) element).getValue()).getImage(element);
			} catch (DebugException e) {
				AcceleoGenDebugUiPlugin.getDefault().log(e, true);
			}
		}
		return null;
	}

	/* (non-Javadoc) */
	public String getText(Object element) {
		String vRet = AcceleoGenDebugUIMessages.getString("AcceleoModelPresentation.ElementText"); //$NON-NLS-1$
		if (element instanceof AcceleoLineBreakpoint) {
			// Gestion d'un point d'arret
			AcceleoLineBreakpoint breakPoint = (AcceleoLineBreakpoint) element;
			try {

				String filePath = breakPoint.getFile() == null ? "" : breakPoint.getFile().getAbsolutePath(); //$NON-NLS-1$

				if (filePath.lastIndexOf(File.separatorChar) != -1) {
					filePath = filePath.substring(filePath.lastIndexOf(File.separatorChar));
				}
				vRet = AcceleoGenDebugUIMessages.getString("AcceleoModelPresentation.BreakpointText", new Object[] { filePath, Integer.toString(breakPoint.getLineNumber()), }); //$NON-NLS-1$
			} catch (CoreException e) {
				AcceleoGenDebugUiPlugin.getDefault().log(e, true);
			}
		} else if (element instanceof AcceleoProcess) {
			vRet = AcceleoGenDebugUIMessages.getString("AcceleoModelPresentation.ProcessText"); //$NON-NLS-1$
		} else if (element instanceof AcceleoThread) {
			vRet = AcceleoGenDebugUIMessages.getString("AcceleoModelPresentation.ThreadText"); //$NON-NLS-1$
		} else if (element instanceof AcceleoDebugTarget) {
			vRet = AcceleoGenDebugUIMessages.getString("AcceleoModelPresentation.TargetText"); //$NON-NLS-1$
		} else if (element instanceof AcceleoStackFrame) {
			AcceleoStackFrame stackFrame = (AcceleoStackFrame) element;
			try {
				vRet = stackFrame.getSourceName() + ' ' + (stackFrame.getLineNumber());
			} catch (DebugException e) {
				AcceleoGenDebugUiPlugin.getDefault().log(e, true);
			}
		}
		return vRet;
	}

	/* (non-Javadoc) */
	public void computeDetail(IValue value, IValueDetailListener listener) {
		String detail = ""; //$NON-NLS-1$
		try {
			detail = value.getValueString();
		} catch (DebugException e) {
		}
		listener.detailComputed(value, detail);
	}

	/* (non-Javadoc) */
	public IEditorInput getEditorInput(Object element) {
		if (element instanceof File) {
			IFile workspaceFile = ResourcesPlugin.getWorkspace().getRoot().getFileForLocation(new Path(((File) element).getAbsolutePath()));
			if (workspaceFile != null && workspaceFile.isAccessible()) {
				return new FileEditorInput(workspaceFile);
			}
		}
		if (element instanceof IFile) {
			return new FileEditorInput((IFile) element);
		}
		if (element instanceof ILineBreakpoint) {
			return new FileEditorInput((IFile) ((ILineBreakpoint) element).getMarker().getResource());
		}
		return null;
	}

	/* (non-Javadoc) */
	public String getEditorId(IEditorInput input, Object element) {
		String name;
		if (element instanceof File) {
			name = ((File) element).getName();
		} else if (element instanceof IFile) {
			name = ((IFile) element).getName();
		} else if (element instanceof ILineBreakpoint) {
			name = ((IFile) ((ILineBreakpoint) element).getMarker().getResource()).getName();
		} else {
			name = ""; //$NON-NLS-1$
		}
		IWorkbench workbench = PlatformUI.getWorkbench();
		IWorkbenchWindow workbenchWindow = workbench.getActiveWorkbenchWindow();
		if (workbenchWindow != null) {
			IWorkbenchPage page = workbenchWindow.getActivePage();
			if (page != null) {
				IEditorDescriptor descriptor = workbench.getEditorRegistry().getDefaultEditor(name);
				if (descriptor != null) {
					return descriptor.getId();
				}
			}
		}
		return getEditorId(name);
	}

	private String getEditorId(String name) {
		String editorID = null;
		String ext = new Path(name).getFileExtension();
		if (ext != null) {
			editorID = (String) ext2editorID.get(ext);
			if (editorID == null) {
				IExtensionRegistry registry = Platform.getExtensionRegistry();
				IExtensionPoint extensionPoint = registry.getExtensionPoint(DEBUG_EDITOR_EXTENSION_ID);
				if (extensionPoint == null || extensionPoint.getExtensions().length == 0) {
					AcceleoGenDebugUiPlugin.getDefault().log(
							AcceleoGenDebugUIMessages.getString("AcceleoModelPresentation.UnresolvedExtensionPoint", new Object[] { DEBUG_EDITOR_EXTENSION_ID, }), true); //$NON-NLS-1$
				} else {
					IExtension[] extensions = extensionPoint.getExtensions();
					for (int i = 0; editorID == null && i < extensions.length; i++) {
						IExtension extension = extensions[i];
						IConfigurationElement[] members = extension.getConfigurationElements();
						for (int j = 0; editorID == null && j < members.length; j++) {
							IConfigurationElement member = members[j];
							String fileExtension = member.getAttribute("extension"); //$NON-NLS-1$
							if (ext.equalsIgnoreCase(fileExtension)) {
								editorID = member.getAttribute("editorID"); //$NON-NLS-1$
							}
						}
					}
				}
				ext2editorID.put(ext, editorID);
			}

		}
		if (editorID != null) {
			return editorID;
		} else {
			return "fr.obeo.acceleo.gen.ui.editors.template.AcceleoTemplateEditor"; //$NON-NLS-1$
		}
	}

	private Map ext2editorID = new HashMap();

}
